package bls

import (
	"fmt"
	mrand "math/rand"
	"testing"

	blst "github.com/supranational/blst/bindings/go"
)

func TestSignMultipleVerifyAggregateMinSig(t *testing.T) {
	msgCount := 5
	for size := 1; size < 20; size++ {
		msgs := make([]Message, 0)
		sks := make([]*SecretKey, 0)
		pks := make([]*PublicKeyMinSig, 0)

		// Generate messages
		for i := 0; i < msgCount; i++ {
			msg := Message(fmt.Sprintf("blst is a blast!! %d %d", i, size))
			msgs = append(msgs, msg)
		}

		// Generate keypairs
		for i := 0; i < size; i++ {
			priv := GenRandomKeyMinSig()
			sks = append(sks, priv)
			pks = append(pks, new(PublicKeyMinSig).From(priv))
		}

		// All signers sign each message
		aggSigs := make([]*SignatureMinSig, 0)
		aggPks := make([]*PublicKeyMinSig, 0)
		for i := 0; i < msgCount; i++ {
			sigsToAgg := make([]*SignatureMinSig, 0)
			pksToAgg := make([]*PublicKeyMinSig, 0)
			for j := 0; j < size; j++ {
				sigsToAgg = append(sigsToAgg,
					new(SignatureMinSig).Sign(sks[j], msgs[i],
						dstMinSig))
				pksToAgg = append(pksToAgg, pks[j])
			}

			agSig := new(AggregateSignatureMinSig)
			if !agSig.Aggregate(sigsToAgg, true) {
				t.Errorf("failed to aggregate")
			}
			afSig := agSig.ToAffine()
			agPk := new(AggregatePublicKeyMinSig)
			agPk.Aggregate(pksToAgg, false)
			afPk := agPk.ToAffine()
			aggSigs = append(aggSigs, afSig)
			aggPks = append(aggPks, afPk)

			// Verify aggregated signature and pk
			if !afSig.Verify(false, afPk, false, msgs[i], dstMinSig) {
				t.Errorf("failed to verify single aggregate size %d", size)
			}

		}

		randFn := func(s *Scalar) {
			var rbytes [blst.BLST_SCALAR_BYTES]byte
			mrand.Read(rbytes[:])
			s.FromBEndian(rbytes[:])
		}

		// Verify
		randBits := 64
		if !new(SignatureMinSig).MultipleAggregateVerify(aggSigs, true,
			aggPks, false,
			msgs, dstMinSig,
			randFn, randBits) {
			t.Errorf("failed to verify multiple aggregate size %d", size)
		}

		// Negative test
		if new(SignatureMinSig).MultipleAggregateVerify(aggSigs, true,
			aggPks, false,
			msgs, dstMinSig[1:],
			randFn, randBits) {
			t.Errorf("failed to not verify multiple aggregate size %d", size)
		}
	}
}
